#include "stdafx.h"
#include "clipboard.hpp"
#include "clipview.hpp"
#include "graphics.hpp"
#include "app.hpp"

#pragma comment(lib, "shlwapi")

#if _DEBUG
#   pragma comment(lib, "libpng/projects/visualc71/Win32_LIB_Debug/libpngd.lib")
#   pragma comment(lib, "libpng/projects/visualc71/Win32_LIB_Debug/ZLib/zlibd.lib")
#else
#   pragma comment(lib, "libpng/projects/visualc71/Win32_LIB_Release/libpng.lib")
#   pragma comment(lib, "libpng/projects/visualc71/Win32_LIB_Release/ZLib/zlib.lib")
#endif

using namespace std;


// allocate statics
HINSTANCE App::inst;
HWND App::mainWnd;
HWND App::nextViewer;
wstring App::outDir;
wstring App::outFmt;
bool App::includeInitial = false;
bool App::verbose = false;
bool App::utcTimes = false;
bool App::silent = false;


BOOL WINAPI ctrlHandler(DWORD type)
{
    if (type == CTRL_C_EVENT || type == CTRL_BREAK_EVENT || type == CTRL_CLOSE_EVENT) {
        if (App::mainWnd) {
            // Send WM_CLOSE to destroy the main window and (most importantly)
            // to remove itself from the clipboard chain. The default handler
            // (called eventually after we return false here) will call
            // ExitProcess. The main thread may not run to completion, 
            // but the clipboard chain is the only thing that MUST be handled 
            // explicitly.
            SendMessage(App::mainWnd, WM_CLOSE, 0, 0);
        }
    }
    return false;  // continue handler
}


void validateOption(string name, wstring value, wstring choices)
{
    std::vector<wstring> v;
    boost::split(v, choices, boost::is_any_of(L","));
    if (std::find(v.begin(), v.end(), value) != v.end())
        return;
    namespace po = boost::program_options;
    throw po::validation_error(
        po::validation_error::invalid_option_value, 
        name);
}


bool parseArgs(int argc, wchar_t* argv[])
{
    bool desktop;

    namespace po = boost::program_options;
    po::positional_options_description pos_desc;
    po::options_description desc("options");
    // for double string literals in default_value
    // see http://stackoverflow.com/questions/6921196
    // ("In Boost::Program_Options, how to set default value for wstring?")
    desc.add_options()
        ("format,f", po::wvalue<wstring>(&App::outFmt)->default_value(L"png", "png"), 
            "output format; one of: bmp, png")
        ("dest,d", po::wvalue<wstring>(&App::outDir)->default_value(L".", "."), 
            "destination directory; will be created if missing")
        ("initial,i", po::bool_switch(&App::includeInitial)->default_value(false),
            "include initial, existing image")
        ("desktop,D", po::bool_switch(&desktop)->default_value(false),
            "use Deskop folder as destination; overrides -d")
        ("utc,u", po::bool_switch(&App::utcTimes)->default_value(false),
            "output UTC times, instead of local; appends \"Z\"")
        ("silent,s", po::bool_switch(&App::silent)->default_value(false),
            "prevent beep on image save")
        ("verbose,v", po::bool_switch(&App::verbose)->default_value(false),
            "verbose output")
        ("help,h", 
            "this help")
        ;

    po::variables_map vm;
    store(
        po::wcommand_line_parser(argc, argv)
            .options(desc)
            .positional(pos_desc)
            .run(),
        vm);
    notify(vm);

    if (vm.count("help")) {
        wcout << L"ClipSnappy v1.0 - save images copied to the clipboard" << endl;
        wcout << L"Copyright 2015 by Elias Fotinis - Freeware - 2015-01-22" << endl;
        wcout << endl;
        cout << desc;
        wcout << endl;
        wcout << L"output file names are \"YYYYMMDD-hhmmssxxx\" (xxx is msec)" << endl;
        return false;
    }

    validateOption("format", App::outFmt, L"bmp,png");

    if (desktop) {
        wchar_t buf[MAX_PATH];
        HRESULT res = SHGetFolderPath(0, CSIDL_DESKTOPDIRECTORY, 0, SHGFP_TYPE_CURRENT, buf);
        if (FAILED(res) || res == S_FALSE)
            throw std::runtime_error("could not get Desktop path");
        App::outDir = buf;
    }

    return true;
}


int wmain(int argc, wchar_t* argv[])
{
    App::inst = GetModuleHandle(0);

    try {
        if (!parseArgs(argc, argv))
            return 0;
    }
    catch (const boost::program_options::error& x) {
        cerr << x.what() << endl;
        return 2;
    }
    catch (const std::runtime_error& x) {
        cerr << x.what() << endl;
        return 1;
    }

    // create output dir if needed
    if (!PathIsDirectory(App::outDir.c_str())) {
        wchar_t buf[MAX_PATH];
        if (!GetFullPathName(App::outDir.c_str(), MAX_PATH, buf, 0)) {
            App::apiError(L"could not create output directory", L"GetFullPathName");
            return 0;
        }
        int res = SHCreateDirectoryEx(0, buf, 0);
        if (res != ERROR_SUCCESS) {
            App::apiError(L"could not create output directory", L"SHCreateDirectoryEx", res);
            return 0;
        }
    }

    if (!SetConsoleCtrlHandler(ctrlHandler, true)) {
        App::apiError(L"init failed", L"SetConsoleCtrlHandler");
        return 0;
    }

    if (ClipboardViewer::registerClass()) {
        App::mainWnd = ClipboardViewer::createWindow();
        if (App::mainWnd) {
            MSG msg;
            BOOL ret;
            while ((ret = GetMessage(&msg, 0,0,0)) != 0) {
                if (ret == -1) {
                    App::apiError(L"message handling failed", L"GetMessage");
                    return 0;
                }
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
    }

    return 0;
}
