#include "stdafx.h"
#include "graphics.hpp"
#include "clipview.hpp"
#include "app.hpp"

using namespace std;


namespace ClipboardViewer {;


LRESULT CALLBACK wndProc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam);


bool registerClass()
{
    WNDCLASS wc = {};
    wc.hInstance = App::inst;
    wc.lpfnWndProc = wndProc;
    wc.lpszClassName = className;
    return RegisterClass(&wc);
}


HWND createWindow()
{
    return CreateWindowEx(
        0, className, className, 
        WS_POPUP, 0,0,0,0, HWND_MESSAGE, 0, App::inst, 0);
}


int wmCreate(HWND wnd)
{
    // SetClipboardViewer doesn't zero the last error on success, so we must
    // do it manually to be able to tell whether the call failed if it returns
    // 0 (the same must be done at the end of WM_DRAWCLIPBOARD, because it's
    // sent before SetClipboardViewer() returns).
    SetLastError(0);
    App::nextViewer = SetClipboardViewer(wnd);
    DWORD err;
    if (!App::nextViewer && (err = GetLastError()) != 0) {
        App::apiError(L"could not setup clipboard viewer", L"SetClipboardViewer", err);
        return -1;
    }
    if (App::verbose) {
        wcout << L"added to clipboard chain; this/next viewer: "
              << wnd << L"/" << App::nextViewer << endl;
    }
    wcout << L"monitoring clipboard; hit Ctrl-C to exit" << endl;
    return 0;
}


void wmDestroy(HWND wnd)
{
    ChangeClipboardChain(wnd, App::nextViewer);
    if (App::verbose)
        wcout << L"removed from clipboard chain" << endl;
    PostQuitMessage(0);
}


wstring getDibCompressionName(DWORD n)
{
    switch (n) {
    case BI_RGB: return L"none";
    case BI_RLE8: return L"RLE8";
    case BI_RLE4: return L"RLE4";
    case BI_BITFIELDS: return L"bitfields";
    case BI_JPEG: return L"JPEG";
    case BI_PNG: return L"PNG";
    default: return L"unknown";
    }
}


wstring getTimeStamp(bool utc)
{
    SYSTEMTIME st;
    if (utc)
        GetSystemTime(&st);
    else
        GetLocalTime(&st);
    wchar_t buf[100];
    wsprintf(buf, L"%04hu%02hu%02hu-%02hu%02hu%02hu%03hu%s", 
        st.wYear, st.wMonth, st.wDay, 
        st.wHour, st.wMinute, st.wSecond, 
        st.wMilliseconds, utc ? L"Z" : L"");
    return buf;
}


wstring getDibInfo(const BITMAPINFOHEADER& bih)
{
    wostringstream s;
    s << bih.biWidth << L"*" << abs(bih.biHeight) << L" ("
        << (bih.biHeight < 0 ? L"top-down" : L"bottom-up") << L"), "
        << bih.biBitCount << L"-bit, compr:" 
        << getDibCompressionName(bih.biCompression);
    if (bih.biCompression == BI_BITFIELDS) {
        DWORD* bf = (DWORD*)(void*)(&bih + 1);
        s << L":" << hex << setfill(L'0') 
            << setw(8) << bf[0] << L"," 
            << setw(8) << bf[1] << L"," 
            << setw(8) << bf[2] << dec;
    }
    return s.str();
}


void processDib(BITMAPINFOHEADER& bih, SIZE_T size)
{
    if (bih.biCompression != BI_RGB && bih.biCompression != BI_BITFIELDS) {
        App::error(L"unsupported bitmap compression");
    }
    else if (bih.biBitCount != 1 && bih.biBitCount != 4 && 
             bih.biBitCount != 8 && bih.biBitCount != 16 && 
             bih.biBitCount != 24 && bih.biBitCount != 32) 
    {
        App::error(L"unsupported bitmap depth");
    }
    else {

        wstring outPath = App::outDir + L"\\" + getTimeStamp(App::utcTimes);

        // --------- BMP ---------
        if (App::outFmt == L"bmp")
        {
            DWORD ticks = GetTickCount();
            BITMAPFILEHEADER bfh = {};
            bfh.bfType = 0x4d42;  // "BM"
            bfh.bfSize = sizeof(bfh) + size;
            bfh.bfOffBits = sizeof(bfh) + sizeof(BITMAPINFOHEADER);

            int paletteEntries = 0;
            if (bih.biClrUsed)
                paletteEntries = bih.biClrUsed;
            else if (bih.biBitCount < 16)  // biClrUsed == 0 implied
                paletteEntries = 1 << bih.biBitCount;

            bfh.bfOffBits += paletteEntries * sizeof(RGBQUAD);
            if (bih.biCompression == BI_BITFIELDS)
                bfh.bfOffBits += 3 * sizeof(DWORD);


            wstring outFname = outPath + L".bmp";
            ofstream ofs(outFname.c_str(), ios::binary);
            ofs.write((const char*)&bfh, sizeof(bfh));
            ofs.write((const char*)&bih, size);
            ofs.close();
            ticks = GetTickCount() - ticks;
            if (ofs) {
                wcout << L"wrote " << outFname << L" in " << 
                    ticks << L"ms" << endl;
                if (!App::silent)
                    MessageBeep(MB_ICONINFORMATION);
            }
            else {
                App::error(L"writing file stream failed");
            }
        }
#ifdef PNG_SUPPORT
        // --------- PNG ---------
        else if (App::outFmt == L"png")
        {
            DWORD ticks = GetTickCount();
            wstring outFname = outPath + L".png";
            try {
                Graphics::writePng(outFname, bih);
                ticks = GetTickCount() - ticks;
                wcout << L"wrote " << outFname << L" in " << 
                    ticks << L"ms" << endl;
                if (!App::silent)
                    MessageBeep(MB_ICONINFORMATION);
            }
            catch (Graphics::PNGError err) {
                App::error(L"writing file failed; PNG error: " + err.msg);
            }
        }
#endif

    }
}


void wmDrawClipboard(HWND wnd, WPARAM wparam, LPARAM lparam)
{
    bool processContents = true;

    static bool firstCall = true;
    if (firstCall) {
        firstCall = false;
        if (!App::includeInitial)
            processContents = false;
    }

    /*
    HWND owner = GetClipboardOwner();
    const int bufLen = MAX_PATH;
    wchar_t nameBuf[bufLen];
    wchar_t classBuf[bufLen];
    GetWindowText(owner, nameBuf, bufLen);
    GetClassName(owner, classBuf, bufLen);
    wcout << L"  owner: " << owner << L" " << classBuf << L" \"" << nameBuf << L"\"" << endl;

    HINSTANCE ownerInst = (HINSTANCE)GetWindowLong(owner, GWL_HINSTANCE);
    if (ownerInst) {
        GetModuleFileName(ownerInst, nameBuf, bufLen);
        wcout << L"  owner module: " << nameBuf << endl;
    }

    vector<UINT> fmts = Clipboard::getFormats();
    wcout << L"  formats: ";
    bool first = true;
    BOOST_FOREACH(UINT fmt, fmts) {
        if (first)
            first = false;
        else
            wcout << L", ";
        wcout << Clipboard::getFormatName(fmt);
    }
    wcout << endl;
    */

    if (processContents) {
        if (!IsClipboardFormatAvailable(CF_DIB)) {
            if (App::verbose)
                wcout << L"non-bitmap detected" << endl;
        }
        else {
            if (!OpenClipboard(wnd)) {
                wcout << L"could not open clipboard" << endl;
            }
            else {
                HANDLE h = GetClipboardData(CF_DIB);
                if (h) {
                    void* mem = GlobalLock(h);
                    if (mem) {
                        BITMAPINFOHEADER& bih = *(BITMAPINFOHEADER*)mem;
                        if (App::verbose)
                            wcout << L"bitmap detected; " << getDibInfo(bih) << endl;
                        processDib(bih, GlobalSize(h));
                        GlobalUnlock(h);
                    }
                }
                CloseClipboard();
            }
        }
    }

    if (App::nextViewer)
        SendMessage(App::nextViewer, WM_DRAWCLIPBOARD, wparam, lparam);
    // must clear the last error, because this message can be sent just before SetClipboardViewer() returns
    SetLastError(0);
}


void wmChangeCbChain(HWND wnd, HWND removed, HWND next)
{
    /*wcout << L"clipboard chain changed" << endl;
    wcout << L"  removed: " << removed << endl;
    wcout << L"  next: " << next << endl;*/
    if (removed == App::nextViewer) {
        App::nextViewer = next;
    }
    else if (App::nextViewer) {
        SendMessage(App::nextViewer, WM_CHANGECBCHAIN, (WPARAM)removed, (LPARAM)next);
    }
}


LRESULT CALLBACK wndProc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
        case WM_CREATE:
            return wmCreate(wnd);
        case WM_DESTROY:
            return wmDestroy(wnd), 0;
        case WM_DRAWCLIPBOARD:
            return wmDrawClipboard(wnd, wparam, lparam), 0;
        case WM_CHANGECBCHAIN:
            return wmChangeCbChain(wnd, (HWND)wparam, (HWND)lparam), 0;
    }
    return DefWindowProc(wnd, msg, wparam, lparam);
}


}  // namespace ClipboardViewer
