#include "stdafx.h"
#include "graphics.hpp"


#ifdef PNG_SUPPORT


void expand565to24(const void* in, void* out, size_t count)
{
    const WORD* src = (const WORD*)in;
    BYTE* dst = (BYTE*)out;
    while (count--) {
        *dst++ = BYTE((*src & 0xf800) >> 8);  // >> 11 << 3
        *dst++ = BYTE((*src & 0x07e0) >> 3);  // >>  5 << 2
        *dst++ = BYTE((*src & 0x001f) << 3);  // >>  0 << 3
        ++src;
    }
}


void expand555to24(const void* in, void* out, size_t count)
{
    const WORD* src = (const WORD*)in;
    BYTE* dst = (BYTE*)out;
    while (count--) {
        *dst++ = BYTE((*src & 0x7c00) >> 7);  // >> 10 << 3
        *dst++ = BYTE((*src & 0x03e0) >> 2);  // >>  5 << 3
        *dst++ = BYTE((*src & 0x001f) << 3);  // >>  0 << 3
        ++src;
    }
}


void shrink24to565(const void* in, void* out, size_t count)
{
    const BYTE* src = (const BYTE*)in;
    WORD* dst = (WORD*)out;
    while (count--) {
        *dst++ = 
            (WORD(src[0] >> 3) << 11) |
            (WORD(src[1] >> 2) << 5) |
            WORD(src[2] >> 3);
        src += 3;
    }
}


void shrink24to555(const void* in, void* out, size_t count)
{
    const BYTE* src = (const BYTE*)in;
    WORD* dst = (WORD*)out;
    while (count--) {
        *dst++ = 
            (WORD(src[0] >> 3) << 10) |
            (WORD(src[1] >> 3) << 5) |
            WORD(src[2] >> 3);
        src += 3;
    }
}


void expand8to24(const void* in, void* out, size_t count, const RGBQUAD* palette)
{
    const BYTE* src = (const BYTE*)in;
    BYTE* dst = (BYTE*)out;
    while (count--) {
        const RGBQUAD& entry = palette[*src++];
        *dst++ = entry.rgbBlue;
        *dst++ = entry.rgbGreen;
        *dst++ = entry.rgbRed;
    }
}


void expand4to24(const void* in, void* out, size_t count, const RGBQUAD* palette)
{
    const BYTE* src = (const BYTE*)in;
    BYTE* dst = (BYTE*)out;
    while (count--) {
        const RGBQUAD& entry1 = palette[*src >> 4];
        *dst++ = entry1.rgbBlue;
        *dst++ = entry1.rgbGreen;
        *dst++ = entry1.rgbRed;
        if (!count--)
            break;
        const RGBQUAD& entry2 = palette[*src++ & 0xf];
        *dst++ = entry2.rgbBlue;
        *dst++ = entry2.rgbGreen;
        *dst++ = entry2.rgbRed;
    }
}


namespace {


    template <typename T>
    std::vector<T*> getDibRowPtrs(const BITMAPINFOHEADER& bih, T* base)
    {
        std::vector<T*> ret;
        bool isUncompressed = bih.biCompression == BI_RGB || bih.biCompression == BI_BITFIELDS;
        if (base && isUncompressed) {
            int width = bih.biWidth;
            int height = abs(bih.biHeight);
            bool bottomUp = bih.biHeight > 0;
            int rowStride = ((bih.biBitCount * width + 31) / 32) * 4;
            char* p = reinterpret_cast<char*>(base);
            for (int i = 0; i < height; ++i) {
                ret.push_back(reinterpret_cast<T*>(p));
                p += rowStride;
            }
            if (bottomUp)
                reverse(ret.begin(), ret.end());
        }
        return ret;
    }


    png_color convertToPng(const RGBQUAD& rgb)
    {
        png_color ret;
        ret.red = rgb.rgbRed;
        ret.green = rgb.rgbGreen;
        ret.blue = rgb.rgbBlue;
        return ret;
    }


    std::vector<png_color> getDibPalette(const RGBQUAD* palette, int size)
    {
        std::vector<png_color> ret;
        std::transform(palette, palette + size, std::back_inserter(ret), convertToPng);
        return ret;
    }


    // 8-8-8 BGR
    bool hasNormal32Bitfields(const BITMAPINFOHEADER& bih)
    {
        const DWORD* n = reinterpret_cast<const DWORD*>(&bih + 1);
        return n[0] == 0x00ff0000 && n[1] == 0x0000ff00 && n[2] == 0x000000ff;
    }


    // 5-6-5 BGR
    bool hasNormal16Bitfields(const BITMAPINFOHEADER& bih)
    {
        const DWORD* n = reinterpret_cast<const DWORD*>(&bih + 1);
        return n[0] == 0x0000f800 && n[1] == 0x000007e0 && n[2] == 0x0000001f;
    }


}


void Graphics::writePng(std::wstring fname, const BITMAPINFOHEADER& bih)
{
    ScopedFILE fp(_wfopen(fname.c_str(), L"wb"));
    if (!fp)
        throw PNGError(L"could not open output file");

    Scoped_png_write_struct png_ptr(
        png_create_write_struct(PNG_LIBPNG_VER_STRING, 0,0,0));
    if (!png_ptr)
        throw PNGError(L"could not allocate png_struct");

    Scoped_png_info_struct info_ptr(
        png_create_info_struct(png_ptr), png_ptr);
    if (!info_ptr)
        throw PNGError(L"could not allocate info_struct");

    if (setjmp(png_jmpbuf(png_ptr)))
        throw PNGError(L"longjmp setup failed");

    png_init_io(png_ptr, fp);
    // set row callback and compression level here, if needed

    int pngBitDepth = bih.biBitCount <= 8 ? bih.biBitCount : 8;
    int pngClrType = bih.biBitCount <= 8 ? PNG_COLOR_TYPE_PALETTE : PNG_COLOR_TYPE_RGB;

    png_set_IHDR(png_ptr, info_ptr, bih.biWidth, abs(bih.biHeight), 
        pngBitDepth, pngClrType, PNG_INTERLACE_NONE, 
        PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    const char* p = reinterpret_cast<const char*>(&bih + 1);
    const DWORD* const bitfields = reinterpret_cast<const DWORD*>(bih.biCompression == BI_BITFIELDS ? p : 0);
    if (bitfields)
        p += 3 * sizeof(DWORD);
    const RGBQUAD* const palette = reinterpret_cast<const RGBQUAD*>(bih.biBitCount <= 8 ? p : 0);
    const int paletteSize = (bih.biBitCount <= 8) ? (1 << bih.biBitCount) : 0;
    if (palette)
        p += paletteSize * sizeof(RGBQUAD);
    void* bits = const_cast<char*>(p);

    std::vector<png_byte*> row_ptrs = getDibRowPtrs(bih, 
        reinterpret_cast<png_byte*>(bits));
    if (row_ptrs.empty())
        throw PNGError(L"could not allocate row pointers");

    if (bih.biBitCount <= 8) {
        std::vector<png_color> pal = getDibPalette(palette, paletteSize);
        png_set_PLTE(png_ptr, info_ptr, &pal[0], pal.size());
        png_set_rows(png_ptr, info_ptr, &row_ptrs.front());
        png_write_png(png_ptr, info_ptr, 0, 0);
        return;
    }
    else if (bih.biBitCount == 16 && hasNormal16Bitfields(bih)) {
        png_write_info(png_ptr, info_ptr);

        std::vector<png_byte> outbuf(bih.biWidth * 3);
        const int imageHeight = bih.biHeight >= 0 ? bih.biHeight : -bih.biHeight;
        for (int y = 0; y < imageHeight; ++y) {
            WORD* src = reinterpret_cast<WORD*>(row_ptrs[y]);
            BYTE* dst = reinterpret_cast<BYTE*>(&outbuf[0]);
            for (int x = 0; x < bih.biWidth; ++x, ++src) {
                *dst++ = BYTE((*src & 0xf800) >> 8);
                *dst++ = BYTE((*src & 0x07e0) >> 3);
                *dst++ = BYTE((*src & 0x001f) << 3);
            }

            png_bytep row_pointer = &outbuf[0];
            png_write_row(png_ptr, row_pointer);
        }

        png_write_end(png_ptr, info_ptr);
        return;
    }
    else {
        png_set_rows(png_ptr, info_ptr, &row_ptrs.front());

        if (bih.biBitCount == 24) {
            png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_BGR, 0);
            return;
        }
        else if (bih.biBitCount == 32) {
            if (bih.biCompression == BI_RGB || 
                bih.biCompression == BI_BITFIELDS && hasNormal32Bitfields(bih))
            {
                png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_BGR | PNG_TRANSFORM_STRIP_FILLER_AFTER, 0);
                return;
            }
        }
    }

    throw PNGError(L"unsupported DIB format");
}


#endif
